==> locus-0.88.tar.gz: locus-0.88/lib/fbase/fbdual.cpp <==
#include "myfndt.h"
#include "fbdual.h"
#include "hdoc.h"
#include "locusopt.h"
#include "dualmap.h"

#include "oflbase.h"
#include "odocfile.h"
#include "obase.h"

//#define DEBUG_notfound

CFBDual::CFBDual(ushort Opt,
	CLocusOptions &Options)
{
	DocFilesAlias = new CDualMap(Opt,
		(ushort)(Options.COptions<int>::Get(O_AVEFILELEN)),
		Options.Get(O_DFNUMTREE),
		Options.Get(O_DFSTRTREE),
		Options.Get(O_DFCONDLIST),
		Options.Get(O_DFCONDDATA));
	DocHandlerAlias = new CDocHandler(DocFilesAlias);

	NumToDoc = new CBook(new CULongHandler(),
		Options.Get(O_BASENUMTREE),
		Opt);

	DocToNum = new CBook(DocHandlerAlias,
		Options.Get(O_BASESTRTREE),
		Opt);

	Conductor = new CConductor(Options.Get(O_BASECONLISTEXT),
		Options.Get(O_BASECONDATAEXT),
		Opt);
}

CFBDual::~CFBDual()
{
	delete Conductor;
	delete DocToNum;
	// DocHandlerAlias is invalid here
	delete NumToDoc;
}

short CFBDual::Add(ulong Num, const CDocDesc &DocDesc)
{   LONG Data;
    short KnownRet;

#if defined(DEBUG_notfound)
    printf("enter CFBDual::Add; %s (from %d) has num %d\n",
        DocDesc.GetFileName().c_str(),
        DocDesc.GetStartOffs(),
        Num);
#endif

    // check duplicities
    short RetVal = 0;

    CNewKey NumKey(CULongHandler::NewKey(Num));
    if (!NumToDoc->GetData(NumKey, Data))
		RetVal |= 1;

	CNewKey DocKey(CDocHandler::NewKey(DocDesc));
	if (!DocToNum->GetData(DocKey, Data))
		RetVal |= 2;

	if (RetVal)
		return RetVal;

	// add
	short Len = DocHandlerAlias->GetDiskLenReal();
	long Ticket = Conductor->Add(Len);
	Conductor->Put(Ticket, 0, (KEY)DocKey, Len);
	KnownRet = NumToDoc->Add(NumKey, Ticket);
	Assert(KnownRet == 0);
	KnownRet = DocToNum->Add(DocKey, Num);
	Assert(KnownRet == 0);
#if defined(DEBUG_notfound)
    CLockedKey Key1(DocToNum->LockFirstKey(), DocToNum);
    while (Key1 != NULL)
    {   long HighLong;
        long LowLong;
        CTwoLongHandler::GetTwoLong(Key1, HighLong, LowLong);
        printf("(%d, %d)\n", HighLong, LowLong);
        CLockedKey Key2(DocToNum->LockNextKey(Key1), DocToNum);
        Key1 = Key2.Drop();
    }
    printf("\n");
#endif
	return 0;
}

short CFBDual::Del(ulong Num)
{	long Ticket;
	short KnownRet;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (NumToDoc->GetData(NumKey, Ticket))
		 return 1;

	CMemPtr<void> Data = Conductor->Alloc(Ticket);
	KEY DocKey = reinterpret_cast<KEY>((void *)Data);
	KnownRet = DocToNum->Del(DocKey);
	Assert(KnownRet == 0);
	Conductor->Del(Ticket);

	KnownRet = NumToDoc->Del(NumKey);
	Assert(KnownRet == 0);

	return 0;
}

short CFBDual::Del(const CDocDesc &DocDesc)
{	LONG Data;
	long Ticket;
	short KnownRet;

	CNewKey DocKey(CDocHandler::NewKey(DocDesc));
	if (DocToNum->GetData(DocKey, Data))
		return 1;

	ulong Num = Data;
	KnownRet = DocToNum->Del(DocKey);
	Assert(!KnownRet);

	CNewKey NumKey(CULongHandler::NewKey(Num));
	KnownRet = NumToDoc->GetData(NumKey, Ticket);
	Assert(KnownRet == 0);
	Conductor->Del(Ticket);

	KnownRet = NumToDoc->Del(NumKey);
	Assert(KnownRet == 0);

	return 0;
}

short CFBDual::Get(ulong Num, CDocDesc &DocDesc)
{	long Ticket;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (NumToDoc->GetData(NumKey, Ticket))
		return 1;

	CMemPtr<void> NewData = Conductor->Alloc(Ticket);
	KEY DocKey = reinterpret_cast<KEY>((void *)NewData);
	CDocHandler::GetDocDesc(DocKey, DocDesc);
	return 0;
}

short CFBDual::Get(const CDocDesc &DocDesc, ulong &Num)
{	long Data;

#if defined(DEBUG_cut_exception)
    printf("CFBDual::Get: find file %s\n", DocDesc.GetFileName().c_str());
#endif
    // if file isn't in filebase, CDocHandler::NewKey tries to add
    // it into FBDual; can't let that happen when FBDual is read-only (and should
    // not let it happen otherwise)
    if (!CDocHandler::Has(DocDesc.GetFileName()))
        return 1;

	CNewKey DocKey(CDocHandler::NewKey(DocDesc));
#if defined(DEBUG_cut_exception)
    printf("CFBDual::Get: constructed doc key\n");
#endif
	if (DocToNum->GetData(DocKey, Data))
    {
#if defined(DEBUG_cut_exception)
        printf("CFBDual::Get: not found\n");
#endif
		return 1;
    }
	else
	{	Num = Data;
#if defined(DEBUG_cut_exception)
        printf("CFBDual::Get: found %ld\n", Num);
#endif
		return 0;
   }
}

void CFBDual::StoreAll()
{
	NumToDoc->StoreAll();
#if defined(DEBUG_notfound)
    CLockedKey Key1(DocToNum->LockFirstKey(), DocToNum);
    while (Key1 != NULL)
    {   long HighLong;
        long LowLong;
        CTwoLongHandler::GetTwoLong(Key1, HighLong, LowLong);
        printf("(%d, %d)\n", HighLong, LowLong);
        CLockedKey Key2(DocToNum->LockNextKey(Key1), DocToNum);
        Key1 = Key2.Drop();
    }
    printf("\n");
#endif
	DocToNum->StoreAll();
	DocHandlerAlias->StoreAll();
	Conductor->StoreAll();
}

short CFBDual::Has(const DString &DocName)
{
	return DocHandlerAlias->Has(DocName);
}
/*{	char *Str;
	long Num;
	short Dummy;

	CNewKey DocKey(CDocHandler::NewKey(DocName.c_str(), 0, 0));
	CLockedKey NextKey(DocToNum->LockKeyAfter(DocKey), DocToNum);
	if (!NextKey)
		return 0;

	CDocHandler::GetDocDesc(NextKey, Str, Num, Dummy);
	Assert(Num == 0);
	return DocName == Str ? 1 : 0;
}*/

==> locus-0.94.tar.gz: locus-0.94/lib/fbase/fbdual.cpp <==
#include "myfndt.h"
#include "fbdual.h"
#include "hdoc.h"
#include "locusopt.h"
#include "dualmap.h"

#include "oflbase.h"
#include "odocfile.h"
#include "obase.h"

//#define DEBUG_notfound

CFBDual::CFBDual(ushort Opt,
    CLocusOptions &Options)
{
    DocFilesAlias = new CDualMap(Opt,
	(ushort)((Options.*(&COptions<int>::Get))(O_AVEFILELEN)),
	Options.Get(O_DFNUMTREE),
	Options.Get(O_DFSTRTREE),
	Options.Get(O_DFCONDLIST),
	Options.Get(O_DFCONDDATA));
    DocHandlerAlias = new CDocHandler(DocFilesAlias);

    NumToDoc = new CBook(new CULongHandler(),
	Options.Get(O_BASENUMTREE),
	Opt);

    DocToNum = new CBook(DocHandlerAlias,
	Options.Get(O_BASESTRTREE),
	Opt);

    Conductor = new CConductor(Options.Get(O_BASECONLISTEXT),
	Options.Get(O_BASECONDATAEXT),
	Opt);
}

CFBDual::~CFBDual()
{
    delete Conductor;
    delete DocToNum;
    // DocHandlerAlias is invalid here
    delete NumToDoc;
}

short CFBDual::Add(ulong Num, const CDocDesc &DocDesc)
{   LONG Data;
    short KnownRet;

#if defined(DEBUG_notfound)
    printf("enter CFBDual::Add; %s (from %d) has num %d\n",
        DocDesc.GetFileName().c_str(),
        DocDesc.GetStartOffs(),
        Num);
#endif

    // check duplicities
    short RetVal = 0;

    CNewKey NumKey(CULongHandler::NewKey(Num));
    if (!NumToDoc->GetData(NumKey, Data))
		RetVal |= 1;

	CNewKey DocKey(CDocHandler::NewKey(DocDesc));
	if (!DocToNum->GetData(DocKey, Data))
		RetVal |= 2;

	if (RetVal)
		return RetVal;

	// add
	short Len = DocHandlerAlias->GetDiskLenReal();
	long Ticket = Conductor->Add(Len);
	Conductor->Put(Ticket, 0, (KEY)DocKey, Len);
	KnownRet = NumToDoc->Add(NumKey, Ticket);
	Assert(KnownRet == 0);
	KnownRet = DocToNum->Add(DocKey, Num);
	Assert(KnownRet == 0);
#if defined(DEBUG_notfound)
    CLockedKey Key1(DocToNum->LockFirstKey(), DocToNum);
    while (Key1 != NULL)
    {   long HighLong;
        long LowLong;
        CTwoLongHandler::GetTwoLong(Key1, HighLong, LowLong);
        printf("(%d, %d)\n", HighLong, LowLong);
        CLockedKey Key2(DocToNum->LockNextKey(Key1), DocToNum);
        Key1 = Key2.Drop();
    }
    printf("\n");
#endif
	return 0;
}

short CFBDual::Del(ulong Num)
{	long Ticket;
	short KnownRet;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (NumToDoc->GetData(NumKey, Ticket))
		 return 1;

	CMemPtr<void> Data = Conductor->Alloc(Ticket);
	KEY DocKey = reinterpret_cast<KEY>((void *)Data);
	KnownRet = DocToNum->Del(DocKey);
	Assert(KnownRet == 0);
	Conductor->Del(Ticket);

	KnownRet = NumToDoc->Del(NumKey);
	Assert(KnownRet == 0);

	return 0;
}

short CFBDual::Del(const CDocDesc &DocDesc)
{	LONG Data;
	long Ticket;
	short KnownRet;

	CNewKey DocKey(CDocHandler::NewKey(DocDesc));
	if (DocToNum->GetData(DocKey, Data))
		return 1;

	ulong Num = Data;
	KnownRet = DocToNum->Del(DocKey);
	Assert(!KnownRet);

	CNewKey NumKey(CULongHandler::NewKey(Num));
	KnownRet = NumToDoc->GetData(NumKey, Ticket);
	Assert(KnownRet == 0);
	Conductor->Del(Ticket);

	KnownRet = NumToDoc->Del(NumKey);
	Assert(KnownRet == 0);

	return 0;
}

short CFBDual::Get(ulong Num, CDocDesc &DocDesc)
{	long Ticket;

	CNewKey NumKey(CULongHandler::NewKey(Num));
	if (NumToDoc->GetData(NumKey, Ticket))
		return 1;

	CMemPtr<void> NewData = Conductor->Alloc(Ticket);
	KEY DocKey = reinterpret_cast<KEY>((void *)NewData);
	CDocHandler::GetDocDesc(DocKey, DocDesc);
	return 0;
}

short CFBDual::Get(const CDocDesc &DocDesc, ulong &Num)
{	long Data;

#if defined(DEBUG_cut_exception)
    printf("CFBDual::Get: find file %s\n", DocDesc.GetFileName().c_str());
#endif
    // if file isn't in filebase, CDocHandler::NewKey tries to add
    // it into FBDual; can't let that happen when FBDual is read-only (and should
    // not let it happen otherwise)
    if (!CDocHandler::Has(DocDesc.GetFileName()))
        return 1;

	CNewKey DocKey(CDocHandler::NewKey(DocDesc));
#if defined(DEBUG_cut_exception)
    printf("CFBDual::Get: constructed doc key\n");
#endif
	if (DocToNum->GetData(DocKey, Data))
    {
#if defined(DEBUG_cut_exception)
        printf("CFBDual::Get: not found\n");
#endif
		return 1;
    }
	else
	{	Num = Data;
#if defined(DEBUG_cut_exception)
        printf("CFBDual::Get: found %ld\n", Num);
#endif
		return 0;
   }
}

void CFBDual::StoreAll()
{
	NumToDoc->StoreAll();
#if defined(DEBUG_notfound)
    CLockedKey Key1(DocToNum->LockFirstKey(), DocToNum);
    while (Key1 != NULL)
    {   long HighLong;
        long LowLong;
        CTwoLongHandler::GetTwoLong(Key1, HighLong, LowLong);
        printf("(%d, %d)\n", HighLong, LowLong);
        CLockedKey Key2(DocToNum->LockNextKey(Key1), DocToNum);
        Key1 = Key2.Drop();
    }
    printf("\n");
#endif
	DocToNum->StoreAll();
	DocHandlerAlias->StoreAll();
	Conductor->StoreAll();
}

short CFBDual::Has(const DString &DocName)
{
	return DocHandlerAlias->Has(DocName);
}
/*{	char *Str;
	long Num;
	short Dummy;

	CNewKey DocKey(CDocHandler::NewKey(DocName.c_str(), 0, 0));
	CLockedKey NextKey(DocToNum->LockKeyAfter(DocKey), DocToNum);
	if (!NextKey)
		return 0;

	CDocHandler::GetDocDesc(NextKey, Str, Num, Dummy);
	Assert(Num == 0);
	return DocName == Str ? 1 : 0;
}*/
